/* thttpd.c - tiny/turbo/throttling HTTP server
**
** Copyright (C) 1995 by Jef Poskanzer <jef@acme.com>.  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/


#include "config.h"
#include "version.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <pwd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "libhttpd.h"
#include "timers.h"
#include "match.h"


static char* argv0;

typedef struct {
    char* pattern;
    long limit;
    long rate;
    time_t last_avg;
    off_t bytes_since_avg;
    int num_sending;
    } throttletab;
static throttletab* throttles;
static int numthrottles, maxthrottles;


typedef struct {
    int state;
#define CS_FREE 0
#define CS_SENDING 1
#define CS_WAITING 2
    httpd_conn hc;
    int tnums[MAXTHROTTLENUMS];		/* throttle indexes */
    int numtnums;
    long limit;
    time_t started_at;
    Timer* idlec_timer;
    Timer* wakeup_timer;
    long wouldblock_delay;
    long so_sndbuf;
    off_t bytes;
    off_t bytes_sent;
    char* file_address;
    } connecttab;
static connecttab* connects;
static int numconnects, maxconnects;


static httpd_server* hs = (httpd_server*) 0;


static void
shut_down( void )
    {
    int cnum;

    for ( cnum = 0; cnum < maxconnects; ++cnum )
	if ( connects[cnum].state != CS_FREE )
	    httpd_close_conn( &connects[cnum].hc );
    if ( hs != (httpd_server*) 0 )
	{
	httpd_terminate( hs );
	hs = (httpd_server*) 0;
	}
    }


static void
handle_kill( int sig )
    {
    shut_down();
    syslog( LOG_NOTICE, "exiting due to signal %d", sig );
    closelog();
    exit( 1 );
    }


static void
usage( void )
    {
    (void) fprintf( stderr,
"usage:  %s [-p port] [-d dir] [-r] [-u user] [-c cgipat] [-t throttles] [-h host]\n",
	argv0 );
    exit( 1 );
    }


static void
read_throttlefile( char* throttlefile )
    {
    FILE* fp;
    char buf[5000];
    char* cp;
    int len;
    char pattern[5000];
    long limit;
    struct timeval tv;

    fp = fopen( throttlefile, "r" );
    if ( fp == (FILE*) 0 )
	{
	syslog( LOG_CRIT, "%.80s - %m", throttlefile );
	perror( throttlefile );
	exit( 1 );
	}
    
    (void) gettimeofday( &tv, (struct timezone*) 0 );

    while ( fgets( buf, sizeof(buf), fp ) != (char*) 0 )
	{
	/* Nuke comments. */
	cp = strchr( buf, '#' );
	if ( cp != (char*) 0 )
	    *cp = '\0';

	/* Nuke trailing whitespace. */
	len = strlen( buf );
	while ( len > 0 &&
		( buf[len-1] == ' ' || buf[len-1] == '\t' ||
		  buf[len-1] == '\n' || buf[len-1] == '\r' ) )
	    buf[--len] = '\0';

	/* Ignore empty lines. */
	if ( len == 0 )
	    continue;

	/* Parse line. */
	if ( sscanf( buf, " %[^ \t] %ld", pattern, &limit ) != 2 )
	    {
	    syslog( LOG_CRIT,
		"unparsable line in %.80s - %.80s", throttlefile, buf );
	    (void) fprintf( stderr,
		"unparsable line in %.80s - %.80s\n", throttlefile, buf );
	    continue;
	    }
	
	/* Nuke any leading slashes in pattern. */
	if ( pattern[0] == '/' )
	    (void) strcpy( pattern, &pattern[1] );
	while ( ( cp = strstr( pattern, "|/" ) ) != (char*) 0 )
	    (void) strcpy( cp + 1, cp + 2 );

	/* Check for room in throttles. */
	if ( numthrottles >= maxthrottles )
	    {
	    if ( maxthrottles == 0 )
		{
		maxthrottles = 100;	/* arbitrary */
		throttles = NEW( throttletab, maxthrottles );
		}
	    else
		{
		maxthrottles *= 2;
		throttles = RENEW( throttles, throttletab, maxthrottles );
		}
	    if ( throttles == (throttletab*) 0 )
		{
		syslog( LOG_CRIT, "out of memory" );
		(void) fprintf( stderr, "out of memory\n" );
		exit( 1 );
		}
	    }

	/* Add to table. */
	throttles[numthrottles].pattern = strdup( pattern );
	if ( throttles[numthrottles].pattern == (char*) 0 )
	    {
	    syslog( LOG_CRIT, "out of memory" );
	    (void) fprintf( stderr, "out of memory\n" );
	    exit( 1 );
	    }
	throttles[numthrottles].limit = limit;
	throttles[numthrottles].rate = 0;
	throttles[numthrottles].last_avg = tv.tv_sec;
	throttles[numthrottles].bytes_since_avg = 0;
	throttles[numthrottles].num_sending = 0;

	++numthrottles;
	}
    (void) fclose( fp );
    }


static int
check_throttles( connecttab* c )
    {
    int tnum;

    c->numtnums = 0;
    c->limit = 1234567890L;
    for ( tnum = 0; tnum < numthrottles && c->numtnums < MAXTHROTTLENUMS;
	  ++tnum )
	if ( match( throttles[tnum].pattern, c->hc.expnfilename ) )
	    {
	    c->tnums[c->numtnums++] = tnum;
	    /* If we're way over the limit, don't even start. */
	    if ( throttles[tnum].rate > throttles[tnum].limit * 3 / 2 )
		return 0;
	    ++throttles[tnum].num_sending;
	    c->limit = MIN(
		c->limit, throttles[tnum].limit / throttles[tnum].num_sending );
	    }
    return 1;
    }


static void
clear_throttles( connecttab* c, struct timeval* tvP )
    {
    int i, tnum;
    time_t elapsed;

    for ( i = 0; i < c->numtnums; ++i )
	{
	tnum = c->tnums[i];
	--throttles[tnum].num_sending;
	throttles[tnum].bytes_since_avg += c->bytes_sent;
	elapsed = tvP->tv_sec - throttles[tnum].last_avg;
	if ( elapsed >= THROTTLE_TIME )
	    {
	    throttles[tnum].rate =
		( throttles[tnum].rate +
		  throttles[tnum].bytes_since_avg / elapsed ) / 2;
	    throttles[tnum].bytes_since_avg = 0;
	    }
	}
    }


static void
clear_connection( connecttab* c, struct timeval* tvP )
    {
    httpd_close_conn( &c->hc );
    clear_throttles( c, tvP );
    if ( c->idlec_timer != (Timer*) 0 )
	tmr_destroy( c->idlec_timer );
    if ( c->wakeup_timer != (Timer*) 0 )
	tmr_destroy( c->wakeup_timer );
    if ( c->file_address != (char*) -1 )
	if ( munmap( c->file_address, c->hc.bytes ) < 0 )
	    syslog( LOG_ERR, "munmap - %m" );
    c->state = CS_FREE;
    --numconnects;
    }


static void
idle_connection( ClientData client_data )
    {
    connecttab* c;
    struct timeval tv;

    c = (connecttab*) client_data.p;
    c->idlec_timer = (Timer*) 0;
    if ( c->state != CS_FREE )
	{
	syslog( LOG_NOTICE,
	    "%.80s connection timed out", inet_ntoa( c->hc.client_addr ) );
	(void) gettimeofday( &tv, (struct timezone*) 0 );
	clear_connection( c, &tv );
	}
    }


static void
wakeup_connection( ClientData client_data )
    {
    connecttab* c;

    c = (connecttab*) client_data.p;
    c->wakeup_timer = (Timer*) 0;
    if ( c->state == CS_WAITING )
	c->state = CS_SENDING;
    }


int
main( int argc, char** argv )
    {
    char* cp;
    int argn;
    int debug;
    int port;
    char* dir;
    int do_chroot;
    struct passwd* pwd;
    uid_t uid;
    gid_t gid;
    char* cgi_pattern;
    char* throttlefile;
    char* hostname;
    char* user;
    char cwd[MAXPATHLEN];
    int nfiles;
    fd_set rfdset;
    fd_set wfdset;
    int cnum;
    connecttab* c;
    int r, sz;
    struct timeval tv;
    time_t elapsed;
    ClientData client_data;

    argv0 = argv[0];

    cp = strrchr( argv0, '/' );
    if ( cp != (char*) 0 )
	++cp;
    else
	cp = argv0;
    openlog( cp, LOG_NDELAY|LOG_PID, LOG_FACILITY );

    (void) signal( SIGINT, handle_kill );
    (void) signal( SIGTERM, handle_kill );
    (void) signal( SIGPIPE, SIG_IGN );		/* get EPIPE instead */

    debug = 0;
    port = DEFAULT_PORT;
    dir = (char*) 0;
#ifdef ALWAYS_CHROOT
    do_chroot = 1;
#else /* ALWAYS_CHROOT */
    do_chroot = 0;
#endif /* ALWAYS_CHROOT */
#ifdef CGI_PATTERN
    cgi_pattern = CGI_PATTERN;
#else /* CGI_PATTERN */
    cgi_pattern = (char*) 0;
#endif /* CGI_PATTERN */
    throttlefile = (char*) 0;
    hostname = (char*) 0;
    user = DEFAULT_USER;
    argn = 1;
    while ( argn < argc && argv[argn][0] == '-' )
	{
	if ( strcmp( argv[argn], "-D" ) == 0 )
	    debug = 1;
	else if ( strcmp( argv[argn], "-p" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    port = atoi( argv[argn] );
	    }
	else if ( strcmp( argv[argn], "-d" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    dir = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-r" ) == 0 )
	    do_chroot = 1;
	else if ( strcmp( argv[argn], "-u" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    user = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-c" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    cgi_pattern = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-t" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    throttlefile = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-h" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    hostname = argv[argn];
	    }
	else
	    usage();
	++argn;
	}
    if ( argn != argc )
	usage();

    /* Check port number. */
    if ( port <= 0 )
	{
	syslog( LOG_CRIT, "illegal port number" );
	(void) fprintf( stderr, "illegal port number\n" );
	exit( 1 );
	}

    /* Throttle file. */
    numthrottles = 0;
    maxthrottles = 0;
    if ( throttlefile != (char*) 0 )
	read_throttlefile( throttlefile );

    /* Figure out uid/gid from user. */
    pwd = getpwnam( user );
    if ( pwd == (struct passwd*) 0 )
	{
	syslog( LOG_CRIT, "unknown user - '%.80s'", user );
	(void) fprintf( stderr, "unknown user - '%s'\n", user );
	exit( 1 );
	}
    uid = pwd->pw_uid;
    gid = pwd->pw_gid;

    /* Switch directories if requested. */
    if ( dir != (char*) 0 )
	{
	if ( chdir( dir ) < 0 )
	    {
	    syslog( LOG_CRIT, "chdir - %m" );
	    perror( "chdir" );
	    exit( 1 );
	    }
	}
#ifdef USE_USER_DIR
    else if ( getuid() == 0 )
	{
	/* No explicit directory was specified, we're root, and the
	** USE_USER_DIR option is set - switch to the specified user's
	** home dir.
	*/
	if ( chdir( pwd->pw_dir ) < 0 )
	    {
	    syslog( LOG_CRIT, "chdir - %m" );
	    perror( "chdir" );
	    exit( 1 );
	    }
	}
#endif /* USE_USER_DIR */
    
    /* Get current directory. */
    (void) getcwd( cwd, sizeof(cwd) );
    if ( cwd[strlen( cwd ) - 1] != '/' )
	(void) strcat( cwd, "/" );

    /* Chroot if requested. */
    if ( do_chroot )
	{
	if ( chroot( cwd ) < 0 )
	    {
	    syslog( LOG_CRIT, "chroot - %m" );
	    perror( "chroot" );
	    exit( 1 );
	    }
	(void) strcpy( cwd, "/" );
	}

    /* Figure out how many file descriptors we can use. */
    nfiles = MIN( httpd_get_nfiles(), FD_SETSIZE );

    if ( ! debug )
	{
	/* We're not going to use stdin stdout or stderr from here on, so close
	** them to save file descriptors.
	*/
	(void) fclose( stdin );
	(void) fclose( stdout );
	(void) fclose( stderr );

	/* Daemonize - make ourselves a subprocess. */
	switch ( fork() )
	    {
	    case 0:
	    break;
	    case -1:
	    syslog( LOG_CRIT, "fork - %m" );
	    exit( 1 );
	    default:
	    exit( 0 );
	    }
	}

    /* Initialize the HTTP layer.  Got to do this before giving up root,
    ** so that we can bind to a privileged port.
    */
    hs = httpd_initialize( hostname, port, cgi_pattern, cwd );
    if ( hs == (httpd_server*) 0 )
	exit( 1 );

    /* If we're root, try to become someone else. */
    if ( getuid() == 0 )
	{
	if ( setgid( gid ) < 0 )
	    {
	    syslog( LOG_CRIT, "setgid - %m" );
	    exit( 1 );
	    }
	if ( setuid( uid ) < 0 )
	    {
	    syslog( LOG_CRIT, "setuid - %m" );
	    exit( 1 );
	    }
	/* Check for unnecessary security exposure. */
	if ( ! do_chroot )
	    syslog(
		LOG_CRIT,
		"started as root without requesting chroot(), warning only" );
	}

    /* Initialize our connections table. */
    maxconnects = nfiles - SPARE_FDS;
    connects = NEW( connecttab, maxconnects );
    if ( connects == (connecttab*) 0 )
	{
	syslog( LOG_CRIT, "out of memory" );
	exit( 1 );
	}
    for ( cnum = 0; cnum < maxconnects; ++cnum )
	{
	connects[cnum].state = CS_FREE;
	connects[cnum].hc.initialized = 0;
	}
    numconnects = 0;

    /* Main loop. */
    (void) gettimeofday( &tv, (struct timezone*) 0 );
    for (;;)
	{
	/* Set up the fdsets. */
	FD_ZERO( &rfdset );
	FD_SET( hs->fd, &rfdset );
	FD_ZERO( &wfdset );
	if ( numconnects > 0 )
	    for ( cnum = 0; cnum < maxconnects; ++cnum )
		if ( connects[cnum].state == CS_SENDING )
		    FD_SET( connects[cnum].hc.fd, &wfdset );
	
	/* Do the select. */
	r = select(
	    nfiles, &rfdset, &wfdset, (fd_set*) 0, tmr_timeout( &tv ) );
	if ( r < 0 )
	    {
	    if ( errno == EINTR )
		continue;	/* try again */
	    syslog( LOG_ERR, "select - %m" );
	    exit( 1 );
	    }
	(void) gettimeofday( &tv, (struct timezone*) 0 );
	if ( r == 0 )
	    {
	    /* No fd's are ready - run the timers. */
	    tmr_run( &tv );
	    continue;
	    }

	/* Is it a new connection? */
	if ( FD_ISSET( hs->fd, &rfdset ) )
	    {
	    /* Is there room in the connection table? */
	    if ( numconnects >= maxconnects )
		{
		/* Out of connection slots.  Run the timers, then the
		** existing connections, and maybe we'll free up a slot
		** by the time we get back here.
		**/
		syslog( LOG_WARNING, "too many connections!" );
		tmr_run( &tv );
		goto handleConnections;
		}
	    /* Find a free connection entry. */
	    for ( cnum = 0; cnum < maxconnects; ++cnum )
		if ( connects[cnum].state == CS_FREE )
		    break;
	    c = &connects[cnum];

	    /* Get the connection. */
	    if ( httpd_get_conn( hs, &connects[cnum].hc ) < 0 )
		continue;
	    c->state = CS_SENDING;
	    c->started_at = tv.tv_sec;
	    c->idlec_timer = (Timer*) 0;
	    c->wakeup_timer = (Timer*) 0;
	    c->wouldblock_delay = 0;
	    c->bytes_sent = 0;
	    c->file_address = (char*) -1;
	    ++numconnects;

	    /* Check the throttle table */
	    if ( ! check_throttles( c ) )
		{
		httpd_send_err(
		    &c->hc, 502, httpd_err502title,
		    httpd_err502form, c->hc.encodedurl );
		clear_connection( c, &tv );
		continue;
		}
	    
	    /* Start the connection going. */
	    if ( httpd_start_request( &c->hc ) < 0 )
		{
		/* Something went wrong.  Close down the connection. */
		clear_connection( c, &tv );
		continue;
		}
	    if ( c->hc.file_fd < 0 )
		{
		/* It's all handled.  Close down the connection. */
		c->bytes_sent = c->hc.bytes;
		clear_connection( c, &tv );
		continue;
		}
	    /* Cool, we have a valid connection and a file to send to it.
	    ** Make an idle timer for it.
	    */
	    client_data.p = c;
	    c->idlec_timer = tmr_create(
		&tv, idle_connection, client_data,
		IDLEC_TIMELIMIT * 1000000L, 0 );
	    /* Map the file into memory. */
	    c->file_address = mmap(
		0, c->hc.bytes, PROT_READ, MAP_SHARED, c->hc.file_fd, 0 );
	    if ( c->file_address == (char*) -1 )
		{
		syslog( LOG_ERR, "mmap - %m" );
		clear_connection( c, &tv );
		continue;
		}
	    (void) close( c->hc.file_fd );
	    c->hc.file_fd = -1;
	    /* Get the socket's send-buffer size. */
	    sz = sizeof(c->so_sndbuf);
	    c->so_sndbuf = 0;
	    if ( getsockopt(
		     c->hc.fd, SOL_SOCKET, SO_SNDBUF,
		     (char*) &c->so_sndbuf, &sz ) < 0 )
		syslog( LOG_ERR, "getsockopt SO_SNDBUF - %m" );
	    /* Make sure it's at least 4096. */
	    if ( c->so_sndbuf < 4096 )
		{
		c->so_sndbuf = 4096;
		if ( setsockopt(
			 c->hc.fd, SOL_SOCKET, SO_SNDBUF,
			 (char*) &c->so_sndbuf, sz ) < 0 )
		    syslog( LOG_ERR, "setsockopt SO_SNDBUF - %m" );
		}
	    /* Set the socket file descriptor to no-delay mode. */
	    if ( fcntl( c->hc.fd, F_SETFL, O_NDELAY ) < 0 )
		syslog( LOG_ERR, "fcntl O_NDELAY - %m" );

	    /* Go around the loop and do another select, rather than dropping
	    ** through and processing existing connections.  New connections
	    ** always get priority.
	    */
	    continue;
	    }
	handleConnections: ;

	/* If we get here there may be writable file descriptors for one or
	** more connections.  Find 'em and service 'em.
	*/
	for ( cnum = 0; cnum < maxconnects; ++cnum )
	    {
	    c = &connects[cnum];
	    if ( c->state == CS_SENDING &&
		 FD_ISSET( c->hc.fd, &wfdset ) )
		{
		/* Write a bufferload. */
		sz = write(
		    c->hc.fd, &c->file_address[c->bytes_sent],
		    MIN( c->so_sndbuf, c->hc.bytes - c->bytes_sent ) );
		if ( sz == 0 ||
		     ( sz < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ) ) )
		    {
		    /* This shouldn't happen, but some kernels, e.g.
		    ** SunOS 4.1.x, are broken and select() says that
		    ** O_NDELAY sockets are always writable even when
		    ** they're actually not.
		    **
		    ** Current workaround is to block sending on this
		    ** socket for a brief adaptively-tuned period.
		    ** Fortunately we already have all the necessary
		    ** blocking code, for use with throttling.
		    */
		    c->wouldblock_delay += MIN_WOULDBLOCK_DELAY;
		    c->state = CS_WAITING;
		    client_data.p = c;
		    c->wakeup_timer = tmr_create(
			&tv, wakeup_connection, client_data,
			c->wouldblock_delay, 0 );
		    continue;
		    }
		if ( sz < 0 )
		    {
		    /* Something went wrong, close this connection.
		    **
		    ** If it's just an EPIPE, don't bother logging, that
		    ** just means the client hung up on us.
		    **
		    ** On some systems, write() occasionally gives an EINVAL.
		    ** Dunno why, something to do with the socket going
		    ** bad.  Anyway, we don't log those either.
		    */
		    if ( errno != EPIPE && errno != EINVAL )
			syslog( LOG_ERR, "write - %m" );
		    clear_connection( c, &tv );
		    continue;
		    }
		/* Ok, we wrote something. */
		tmr_reset( &tv, c->idlec_timer );
		c->bytes_sent += sz;
		if ( c->bytes_sent >= c->hc.bytes )
		    {
		    /* This conection is finished! */
		    clear_connection( c, &tv );
		    continue;
		    }
		if ( c->wouldblock_delay > MIN_WOULDBLOCK_DELAY )
		    c->wouldblock_delay -= MIN_WOULDBLOCK_DELAY;
		/* Check if we're sending too fast. */
		elapsed = tv.tv_sec - c->started_at;
		if ( elapsed > 0 && c->bytes_sent / elapsed > c->limit )
		    {
		    c->state = CS_WAITING;
		    /* !!! For now just sleep 5 seconds.  Later figure a
		    ** more accurate method.
		    */
		    client_data.p = c;
		    c->wakeup_timer = tmr_create(
			&tv, wakeup_connection, client_data, 5 * 1000000L, 0 );
		    }
		}
	    }
	}
    }
